<sly-select-team-workspace
  :team-id.sync="state.{{{widget.widget_id}}}.teamId"
  :options="data.{{{widget.widget_id}}}.options"
  :disabled="data.{{{widget.widget_id}}}.disabled"
></sly-select-team-workspace>
